import os
import yaml
import gzip
import shutil
import tarfile
import urllib.request
from pathlib import Path

from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns, save_tables

from gretel_client import configure_session
from gretel_client import create_or_get_unique_project
from gretel_client.config import get_session_config
from gretel_client.rest_v1.api.connections_api import ConnectionsApi
from gretel_client.rest_v1.api.workflows_api import WorkflowsApi
from gretel_client.rest_v1.models import (
    CreateWorkflowRunRequest,
    CreateWorkflowRequest,
)
from gretel_client.workflows.logs import print_logs_for_workflow_run

import argparse

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="Biodegradability_v1")
args.add_argument("--real-data-path", type=str, default="data/original")
args.add_argument("--synthetic-data-path", type=str, default="data/synthetic")
args.add_argument("--connection-uid", type=str, required=True)
args.add_argument("--model", type=str, required=True, choices=["lstm", "actgan"])
args.add_argument("--run-id", type=str, default="1")
args = args.parse_args()

dataset_name = args.dataset_name
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
input_connection_uid = args.connection_uid
model = args.model
run_id = args.run_id

# Helpers for running workflows from the notebook


def run_workflow(config: str):
    """Create a workflow, and workflow run from a given yaml config. Blocks and
    prints log lines until the workflow reaches a terminal state.

    Args:
        config: The workflow config to run.
    """
    print("Validating actions in the config...")
    config_dict = yaml.safe_load(config)

    for action in config_dict["actions"]:
        print(f"Validating action {action['name']}")
        response = workflow_api.validate_workflow_action(action)
        print(f"Validation response: {response}")

    workflow = workflow_api.create_workflow(
        CreateWorkflowRequest(
            project_id=project.project_guid,
            config=config_dict,
            name=config_dict["name"],
        )
    )

    workflow_run = workflow_api.create_workflow_run(
        CreateWorkflowRunRequest(workflow_id=workflow.id)
    )

    print(f"workflow: {workflow.id}")
    print(f"workflow run id: {workflow_run.id}")

    print_logs_for_workflow_run(workflow_run.id, session)


# Log into Gretel
configure_session(api_key="prompt", cache="yes", validate=True)


# Load real data
metadata = Metadata().load_from_json(
    Path(real_data_path) / f"{dataset_name}/metadata.json"
)
real_data = load_tables(Path(real_data_path) / f"{dataset_name}", metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
metadata.validate_data(real_data)

# ## Designate Project for your Relational Workflow
table_names = list(real_data.keys())
dataset_name_gretel = dataset_name.replace("_", "-")

session = get_session_config()
connection_api = session.get_v1_api(ConnectionsApi)
workflow_api = session.get_v1_api(WorkflowsApi)

project = create_or_get_unique_project(name=f"Synthesize-{dataset_name_gretel}-{model}")

# Configure and Run your Relational Workflow
# Gretel Workflows provide an easy to use, config driven API for automating and operationalizing synthetic data. A Gretel Workflow is constructed by actions that are composed to create a pipeline for processing data with Gretel. To learn more about Gretel Workflows, check out [our docs](https://docs.gretel.ai/reference/workflows).

### Define Source Data via Connector
connection_type = connection_api.get_connection(input_connection_uid).dict()["type"]

# ### Define Workflow configuration
workflow_config = f"""\
name: {dataset_name_gretel}-{connection_type}-workflow-{model}

actions:
  - name: {connection_type}-read
    type: {connection_type}_source
    connection: {input_connection_uid}
    config:
      sync:
        mode: full

  - name: model-train-run
    type: gretel_tabular
    input: {connection_type}-read
    config:
      project_id: {project.project_guid}
      train:
        model: "synthetics/tabular-{model}"
        dataset: "{{{connection_type}-read.outputs.dataset}}"
      run:
        num_records_multiplier: 1.0

"""
print(workflow_config)


# ### Run Workflow

run_workflow(workflow_config)

# ## View Results

path = f"./data_{dataset_name}_{model}"
os.makedirs(path, exist_ok=True)

# Download the output artifacts
urllib.request.urlretrieve(
    project.get_artifact_link(project.artifacts[-1]["key"]),
    f"./data_{dataset_name}_{model}/workflow-output.tar.gz",
)

with gzip.open(f"./data_{dataset_name}_{model}/workflow-output.tar.gz", "rb") as f_in:
    with open(f"./data_{dataset_name}_{model}/workflow-output.tar", "wb") as f_out:
        f_out.write(f_in.read())

with tarfile.open(f"{path}/workflow-output.tar") as tar:
    tar.extractall(path)


path_synthetic = (
    f"{synthetic_data_path}/{dataset_name}/GRETEL_{model.upper()}/{run_id}/sample1"
)
os.makedirs(path_synthetic, exist_ok=True)
for table in table_names:
    shutil.copy(f"{path}/synth_{table}.csv", f"{path_synthetic}/{table}.csv")


# Postprocess synthetic data (some categorical columns are generated as floats)
def is_float(value):
    if value is None or value == "":
        return False
    try:
        float(value)
        return True
    except:  # noqa: E722
        return False


synthetic_data = load_tables(Path(path_synthetic), metadata)
metadata.validate_data(synthetic_data)

for table in metadata.get_tables():
    table_meta = metadata.get_table_meta(table)
    for column, column_info in table_meta["columns"].items():
        if column_info["sdtype"] == "categorical":
            values = synthetic_data[table][column].unique()
            numeric = False
            for value in values:
                if is_float(str(value)) and value == value:
                    numeric = True
                    break
            if numeric:
                synthetic_data[table][column] = synthetic_data[table][column].astype(
                    "object"
                )
                for i, value in synthetic_data[table][column].items():
                    if value != value or not is_float(str(value)):
                        # skip NaN
                        continue
                    synthetic_data[table].at[i, column] = int(float(value))
                synthetic_data[table][column] = synthetic_data[table][column].astype(
                    "object"
                )

# Save synthetic data
save_tables(synthetic_data, Path(path_synthetic))



import numpy as np
from pathlib import Path

COLORMAP = [
    "#377eb8",
    "#ff7f00",
    "#4daf4a",
    "#f781bf",
    "#a65628",
    "#984ea3",
    "#999999",
    "#e41a1c",
    "#dede00",
]

COLOR_DICT = {
    "SDV": COLORMAP[0],
    "RCTGAN": COLORMAP[1],
    "REALTABFORMER": COLORMAP[2],
    "REALTABF.": COLORMAP[2],
    "MOSTLYAI": COLORMAP[8],
    "GRE-ACTGAN": COLORMAP[3],
    "GRETEL_ACTGAN": COLORMAP[3],
    "GRE-LSTM": COLORMAP[5],
    "GRETEL_LSTM": COLORMAP[5],
    "REAL": COLORMAP[6],
    "CLAVADDPM": COLORMAP[4],
}


def get_color(method_name):
    if method_name in COLOR_DICT:
        return COLOR_DICT[method_name]
    return None


def get_x_tick_width_coef(N):
    return (N - 1) * 0.5


def get_bins(data):
    if data.dtype.name == "category" or data.dtype.name == "object":
        return len(data.unique())
    if data.dtype.name == "bool":
        return 2
    if data.dtype.name == "datetime64" or data.dtype.name == "datetime64[ns]":
        return "auto"
    return np.histogram_bin_edges(data.dropna())


def prettify_metric_name(metric_name):
    if metric_name == "WassersteinDistance":
        return "Wasserstein Distance"
    if metric_name == "KSTest":
        return "Kolmogorov-Smirnov Test"
    if metric_name == "ChiSquareTest":
        return "Chi-Square Test"
    if metric_name == "JSDivergence":
        return "Jensen-Shannon Divergence"
    if metric_name == "MaximumMeanDiscrepancy":
        return "Maximum Mean Discrepancy"
    if metric_name == "PairwiseCorrelationDifference":
        return "Pairwise Correlation Difference"
    return metric_name


def prettify_dataset_name(dataset_name):
    if dataset_name in ("rossmann", "rossmann_store_sales", "rossmann_subsampled"):
        return "Rossmann"
    if dataset_name in ("airbnb-simplified_subsampled", "airbnb-simplified"):
        return "Airbnb"
    if dataset_name in ("Biodegradability_v1"):
        return "Biodegradability"
    if dataset_name in ("Cora_v1"):
        return "Cora"
    if dataset_name in ("imdb_MovieLens_v1"):
        return "IMDB MovieLens"
    if dataset_name in ("walmart", "walmart_subsampled"):
        return "Walmart"
    return dataset_name


def prettify_method_name(method_name):
    if method_name == "REALTABFORMER":
        return "REALTABF"
    if method_name == "GRETEL_ACTGAN":
        return "G-ACTGAN"
    if method_name == "GRETEL_LSTM":
        return "G-LSTM"
    return method_name


def get_dataset_info(
    granularity_level, metric_type, all_results, dataset, methods, **kwargs
):
    base_metrics = list(
        all_results[dataset][list(all_results[dataset].keys())[0]][
            f"{granularity_level}_metrics"
        ].keys()
    )

    if granularity_level == "single_table":
        if metric_type == "distance":
            base_metrics = [
                metric for metric in base_metrics if "detection" not in metric.lower()
            ]
        elif metric_type == "detection":
            base_metrics = [
                metric for metric in base_metrics if "detection" in metric.lower()
            ]
        else:
            raise ValueError(
                f"Unknown metric type {metric_type}. Should be either 'distance' or 'detection'."
            )
    elif granularity_level == "single_column":
        if metric_type == "distance":
            base_metrics = [
                metric for metric in base_metrics if "detection" not in metric.lower()
            ]
            base_metrics = [
                metric for metric in base_metrics if "test" not in metric.lower()
            ]
        elif metric_type == "detection":
            base_metrics = [
                metric for metric in base_metrics if "detection" in metric.lower()
            ]
        else:
            raise ValueError(
                f"Unknown metric type {metric_type}. Should be either 'distance' or 'detection'."
            )
    else:
        raise ValueError(
            f"Unknown granularity level {granularity_level}. Should be either 'single_table' or 'single_column'."
        )

    base_metric_names = [prettify_metric_name(metric) for metric in base_metrics]

    save_figs = kwargs.get("save_figs", False)
    save_figs_path = kwargs.get("save_figs_path", "./figs")
    save_figs_path = Path(save_figs_path) / granularity_level / metric_type

    method_order = kwargs.get(
        "method_order",
        [
            "SDV",
            "RCTGAN",
            "REALTABFORMER",
            "MOSTLYAI",
            "GRETEL_ACTGAN",
            "GRETEL_LSTM",
            "CLAVADDPM",
        ],
    )

    if method_order is not None:
        methods = [
            method
            for method in method_order
            if method in methods and method in all_results[dataset]
        ]
        methods += sorted([method for method in methods if method not in method_order])

    return base_metrics, base_metric_names, save_figs, save_figs_path, methods



"""Dataset Generators for Personal Identifiable Information transformers."""

from abc import ABC

import numpy as np

from rctgan.rdt2.performance.datasets.base import BaseDatasetGenerator
from rctgan.rdt2.performance.datasets.utils import add_nans


class PIIGenerator(BaseDatasetGenerator, ABC):
    """Base class for generators that generate PII data."""

    SDTYPE = 'pii'


class RandomStringGenerator(PIIGenerator):
    """Generator that creates an array of random strings."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        categories = ['Alice', 'Bob', 'Charlie', 'Dave', 'Eve']
        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 500.0
            },
            'transform': {
                'time': 1e-05,
                'memory': 500.0
            },
            'reverse_transform': {
                'time': 2e-05,
                'memory': 1000.0,
            }
        }


class RandomStringNaNsGenerator(PIIGenerator):
    """Generator that creates an array of random strings with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(RandomStringGenerator.generate(num_rows).astype('O'))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 1e-05,
                'memory': 1000.0
            },
            'reverse_transform': {
                'time': 2e-05,
                'memory': 1000.0,
            }
        }



"""SDV Constraints module."""

from rctgan.constraints.base import Constraint
from rctgan.constraints.tabular import (
    Between, ColumnFormula, CustomConstraint, GreaterThan, Negative, OneHotEncoding, Positive,
    Rounding, Unique, UniqueCombinations)

__all__ = [
    'Constraint',
    'ColumnFormula',
    'CustomConstraint',
    'GreaterThan',
    'UniqueCombinations',
    'Between',
    'Negative',
    'Positive',
    'Rounding',
    'OneHotEncoding',
    'Unique'
]
